import ast
import copy
import sys

import yaml

from . import _writer
from .utils import deep_merge, get_by_path, set_by_path


class TrainingConfig:

    def __init__(self, config_files):
        self._config = {}

        # read in all config files
        for file in config_files:
            self._add_config(file)

        self._parse_cmd_args()
        self._raw_config = copy.deepcopy(self._config)

    def get_raw_config(self):
        return self._raw_config

    def _add_config(self, path):
        with open(path) as fh:
            yaml_dict = yaml.load(fh.read(), Loader=yaml.FullLoader)
        self._config = deep_merge(self._config, yaml_dict)

    def _parse_cmd_args(self):
        try:
            i = sys.argv.index('with')
        except ValueError:
            # no config options passed
            return

        config_options = sys.argv[(i + 1):]
        if len(config_options) == 0:
            return

        overwritten = []

        # parse each config option individually
        for option in config_options:
            before, match, after = option.partition('=')
            if match == '=':
                try:
                    value = ast.literal_eval(after)
                except (ValueError, SyntaxError):
                    value = after  # interpret as string
                if before in self:
                    overwritten.append(before)
                set_by_path(self._config, before, value)
            else:
                # named config
                raise NotImplementedError

        if len(overwritten):
            _writer.info(f'Overwriting configuration{"s" if len(overwritten) > 1 else ""} {", ".join(overwritten)} '
                         f'from command line arguments')

    def __contains__(self, item):
        try:
            _ = self[item]
            return True
        except KeyError:
            return False

    def __getitem__(self, item):
        return get_by_path(self._config, item)

    def items(self):
        return self._config.items()

    def __repr__(self):
        return f'{__class__.__name__}(' + ', '.join([f'{k}={repr(v)}' for k, v in self._config.items()]) + ')'



import torch

from .base_backend import BaseBackend


class CPUBackend(BaseBackend):

    def setup(self, model, config_optimizers_fn, checkpoint):
        # setup optimizers for model parameters and return
        return config_optimizers_fn(model=model)

    def get_name(self):
        return f'CPUBackend'

    def train_step(self, train_fn, **train_fn_args):
        return train_fn(**train_fn_args)

    def val_step(self, val_fn, **val_fn_args):
        return val_fn(**val_fn_args)

    def optim_step(self, tensor):
        raise NotImplementedError()

    def scheduler_step(self, metrics):
        for scheduler in self.schedulers.values():
            if isinstance(scheduler, torch.optim.lr_scheduler.ReduceLROnPlateau):
                scheduler.step(metrics)
            else:
                scheduler.step()

    def to_device(self, data):
        return data



# Copyright The PyTorch Lightning team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import importlib
from abc import ABC
from collections.abc import Mapping, Sequence
from copy import copy
from typing import Any, Callable, Union

import torch

TORCHTEXT_AVAILABLE = importlib.util.find_spec("torchtext") is not None
if TORCHTEXT_AVAILABLE:
    from torchtext.data import Batch
else:
    Batch = type(None)


def apply_to_collection(data: Any, dtype: Union[type, tuple], function: Callable, *args, **kwargs) -> Any:
    """
    Recursively applies a function to all elements of a certain dtype.

    Args:
        data: the collection to apply the function to
        dtype: the given function will be applied to all elements of this dtype
        function: the function to apply
        *args: positional arguments (will be forwarded to calls of ``function``)
        **kwargs: keyword arguments (will be forwarded to calls of ``function``)

    Returns:
        the resulting collection

    """
    elem_type = type(data)

    # Breaking condition
    if isinstance(data, dtype):
        return function(data, *args, **kwargs)

    # Recursively apply to collection items
    elif isinstance(data, Mapping):
        return elem_type({k: apply_to_collection(v, dtype, function, *args, **kwargs)
                          for k, v in data.items()})
    elif isinstance(data, tuple) and hasattr(data, '_fields'):  # named tuple
        return elem_type(*(apply_to_collection(d, dtype, function, *args, **kwargs) for d in data))
    elif isinstance(data, Sequence) and not isinstance(data, str):
        return elem_type([apply_to_collection(d, dtype, function, *args, **kwargs) for d in data])

    # data is neither of dtype, nor a collection
    return data


class TransferableDataType(ABC):
    """
    A custom type for data that can be moved to a torch device via `.to(...)`.

    Example:

        >>> isinstance(dict, TransferableDataType)
        False
        >>> isinstance(torch.rand(2, 3), TransferableDataType)
        True
        >>> class CustomObject:
        ...     def __init__(self):
        ...         self.x = torch.rand(2, 2)
        ...     def to(self, device):
        ...         self.x = self.x.to(device)
        ...         return self
        >>> isinstance(CustomObject(), TransferableDataType)
        True
    """

    @classmethod
    def __subclasshook__(cls, subclass):
        if cls is TransferableDataType:
            to = getattr(subclass, "to", None)
            return callable(to)
        return NotImplemented


def move_data_to_device(batch: Any, device: torch.device):
    """
    Transfers a collection of data to the given device. Any object that defines a method
    ``to(device)`` will be moved and all other objects in the collection will be left untouched.

    Args:
        batch: A tensor or collection of tensors or anything that has a method `.to(...)`.
            See :func:`apply_to_collection` for a list of supported collection types.
        device: The device to which the data should be moved

    Return:
        the same collection but with all contained tensors residing on the new device.

    See Also:
        - :meth:`torch.Tensor.to`
        - :class:`torch.device`
    """

    def batch_to(data):
        # try to move torchtext data first
        if TORCHTEXT_AVAILABLE and isinstance(data, Batch):

            # Shallow copy because each Batch has a reference to Dataset which contains all examples
            device_data = copy(data)
            for field in data.fields:
                device_field = move_data_to_device(getattr(data, field), device)
                setattr(device_data, field, device_field)
            return device_data

        kwargs = dict(non_blocking=True) if isinstance(data, torch.Tensor) else {}
        return data.to(device, **kwargs)

    dtype = (TransferableDataType, Batch) if TORCHTEXT_AVAILABLE else TransferableDataType
    return apply_to_collection(batch, dtype=dtype, function=batch_to)



import torch
from torch import cuda

from .base_backend import BaseBackend
from .apply_func import move_data_to_device
from ..utils import AMP_AVAILABLE


class GPUBackend(BaseBackend):

    def __init__(self, gpu_id, enable_amp=False):
        self.gpu_id = gpu_id
        self.enable_amp = enable_amp
        self.device = torch.device(f'cuda:{self.gpu_id}')
        self.model = None
        self.optimizers = {}
        self.schedulers = {}

        if not cuda.is_available() or not cuda.device_count() > self.gpu_id:
            raise ValueError(f'GPU [{self.gpu_id}] is not available.')

        if enable_amp and not AMP_AVAILABLE:
            raise ValueError('AMP is not supported by your PyTorch version, try torch>=1.6.')

    def setup(self, model, config_optimizers_fn, checkpoint=None):
        self.model = model

        if checkpoint:
            self.model.load_state_dict(checkpoint['model'])

        # move model parameters to specified GPU
        self.model.cuda(device=self.gpu_id)

        # setup AMP
        # writer.info('Using AMP with 16 bit precision')
        # TODO setup amp

        # setup optimizers for model parameters
        optimizer_config = config_optimizers_fn(model=self.model)
        self.optimizers = optimizer_config['optimizers']
        self.schedulers = optimizer_config['schedulers']
        if not isinstance(self.optimizers, dict):
            self.optimizers = {'main': self.optimizers}
        if not isinstance(self.schedulers, dict):
            self.schedulers = {'main': self.schedulers}

        if checkpoint:
            # restore optimizer/scheduler states
            for name, state in checkpoint['optimizers'].items():
                self.optimizers[name].load_state_dict(state)
            for name, state in checkpoint['schedulers'].items():
                self.schedulers[name].load_state_dict(state)

    def get_name(self):
        return f'GPUBackend[device={self.gpu_id}, AMP={self.enable_amp}]'

    def train_step(self, train_fn, **train_fn_args):
        if self.enable_amp:
            with cuda.amp.autocast():
                return train_fn(**train_fn_args)
        else:
            return train_fn(**train_fn_args)

    def val_step(self, val_fn, **val_fn_args):
        return val_fn(**val_fn_args)

    def optim_step(self, tensor: torch.Tensor):
        for optimizer in self.optimizers.values():
            optimizer.zero_grad()
            tensor.backward()
            optimizer.step()

    def scheduler_step(self, metrics):
        for scheduler in self.schedulers.values():
            if isinstance(scheduler, torch.optim.lr_scheduler.ReduceLROnPlateau):
                scheduler.step(metrics)
            else:
                scheduler.step()

    def to_device(self, data):
        return move_data_to_device(data, torch.device(self.gpu_id))



import sys
import time
import random
from functools import wraps
from pathlib import Path

from halo import Halo
from tqdm import tqdm
import torch
import numpy as np


AMP_AVAILABLE = hasattr(torch.cuda, "amp") and hasattr(torch.cuda.amp, "autocast")
IS_TTY = sys.stdout.isatty()
NON_TTY_UPDATE_INTERVAL = 8000  # milliseconds


def make_wrapper(f):
	@wraps(f)
	def wrapper(*args, **kwargs):
		return f(*args, **kwargs)
	return wrapper


def deep_merge(a, b):
	"""
	Deep-merges two dicts, with b overwriting values in a.
	https://stackoverflow.com/a/20666342
	"""
	for key, value in b.items():
		if isinstance(value, dict):
			node = a.setdefault(key, {})
			deep_merge(node, value)
		else:
			a[key] = value
	return a


def get_by_path(dic, path):
	"""
	Gets a key (can be specified via dot notation `path`) in a nested `dic`
	"""
	parts = path.split('.')
	loc = dic
	for part in parts:
		if part not in loc:
			raise KeyError(path)
		loc = loc[part]
	return loc


def set_by_path(dic, path, value):
	"""
	Sets a key (can be specified via dot notation `path`) in a nested `dic` to `value`
	"""
	parts = path.split('.')
	loc = dic
	for part in parts[:-1]:
		if part not in loc:
			loc[part] = {}
		loc = loc[part]

	loc[parts[-1]] = value


class _TQDM(tqdm):
	"""
	Provides a subclass of tqdm which returns current format dict to a callback instead of printing it.
	"""

	def __init__(self, iterable, callback=None):
		self.callback = callback
		super().__init__(iterable)

	def display(self, msg=None, pos=None):
		self.callback(self.format_dict)

	def close(self):
		pass


class _Spinner:

	def __init__(self, text):
		self._text = text
		# if we're not in a tty, increase update interval to reduce log file sizes
		self._halo = Halo(text, interval=(-1 if IS_TTY else NON_TTY_UPDATE_INTERVAL))
		self._current_metrics = None

	def __enter__(self):
		self._halo.start()
		return self

	def __exit__(self, exc_type, exc_val, exc_tb):
		if self._halo._stop_spinner.is_set():
			return

		if exc_type is None:
			self._halo.succeed()
		else:
			self._halo.fail()

		time.sleep(0.1)
		self._halo.stop()

	def tqdm(self, iterable):
		orig_text = self._halo.text

		def callback(f_dic):
			# format: Training epoch 2 (batch 1233/56990, 10:00 elapsed, 11:31 left)
			remaining = (f_dic['total'] - f_dic['n']) / f_dic['rate'] if f_dic['rate'] and f_dic['total'] else 0
			remaining_str = tqdm.format_interval(remaining) if f_dic['rate'] else '?'
			elapsed_str = tqdm.format_interval(f_dic["elapsed"])
			if self._current_metrics is not None:
				metrics_text = ''.join(f'{k}: {v}, ' for k, v in self._current_metrics.items())
			else:
				metrics_text = ''
			self._halo.text = \
				f'{orig_text} ({metrics_text}' \
				f'batch {f_dic["n"] + 1}/{f_dic["total"]}, {elapsed_str} elapsed, {remaining_str} left)'

		return _TQDM(iterable, callback=callback)

	def set_current_metrics(self, metric_dict):
		self._current_metrics = metric_dict

	def info(self, text=None):
		self._halo.info(text)

	def warning(self, text=None):
		if text is None:
			text = self._text
		self._halo.warn(text)

	def success(self, text=None):
		if text is None:
			text = self._text
		self._halo.succeed(text)

	def error(self, text=None):
		if text is None:
			text = self._text
		self._halo.fail(text)


class _TerminalWriter:

	@staticmethod
	def task(text):
		return _Spinner(text)

	@staticmethod
	def info(text):
		Halo(text).info()

	@staticmethod
	def warning(text):
		Halo(text).warn()

	@staticmethod
	def success(text):
		Halo(text).succeed()

	@staticmethod
	def error(text):
		Halo(text).fail()


_logger = None


def get_terminal_writer():
	global _logger

	if _logger is not None:
		return _logger

	_logger = _TerminalWriter()
	return _logger


def get_highest_run(save_path: Path):
	highest = 0
	for path in save_path.iterdir():
		num = path.name.split('-')[-1]
		try:
			if int(num) > highest:
				highest = int(num)
		except:
			pass

	return highest


def std_round(value):
	return round(value, 4)


def seed_all(seed):
	random.seed(seed)
	np.random.seed(seed)
	torch.manual_seed(seed)


def set_deterministic(deterministic: bool):
	if torch.cuda.is_available():
		torch.backends.cudnn.benchmark = not deterministic
		torch.backends.cudnn.deterministic = deterministic



import warnings
from datetime import datetime
from typing import Optional, List, Union
from pathlib import Path
import random
import functools

import numpy as np
import torch
from torch.utils.data import DataLoader
from coolname import generate_slug, replace_random

from . import _writer as writer
from .backends.cpu_backend import CPUBackend
from .backends.gpu_backend import GPUBackend
from .utils import make_wrapper, get_highest_run, std_round, seed_all, set_deterministic
from .config import TrainingConfig
from .bound_functions import BoundFunctions
from .loggers import BaseLogger, LoggerSet, StandardLogger


class Run:
    """
    Represents an individual training run.
    """

    def __init__(self, config_files: Optional[List], random_seed: int = None):
        self._bound_functions = BoundFunctions()
        self._config = None
        self._backend = None
        self._model = None
        self._logger = None
        self._max_epochs = None
        self._use_gpu = None
        self._gpu_id = None
        self._resume_checkpoint = None
        self._save_path = None
        self._run_name = None
        self._optimize_metric = None
        self._checkpoint_metric = None
        self._checkpoint_every = None
        self._smaller_is_better = None  # TODO state which to minimize/checkpoint on in result dict
        self._optimize_first = None
        self._enable_amp = None
        self._is_validate = None
        self._config_files = []

        # TODO support run() args in config
        if config_files is None:
            config_files = []
        for file in config_files:
            self.add_config(file)
        self._config = TrainingConfig(self._config_files)

        if random_seed:
            seed_all(random_seed)

    # TODO types, docstrings
    # TODO pin_memory
    # todo clear cache before start
    # todo hooks
    # todo cleanup code (extra files for optim, devices etc.)
    # todo save on ctrl-C
    # todo look at pl GPUbackend (amp optimizuation etc)
    def run(self,
            model: torch.nn.Module,
            train_loader: DataLoader,
            val_loader: DataLoader,
            loggers: Optional[List[BaseLogger]] = None,
            max_epochs=1,
            use_gpu=True,
            gpu_id=0,
            resume_checkpoint: Optional[Union[str, Path]] = None,
            save_path='train_logs',
            run_name=None,
            optimize_metric=None,
            checkpoint_metric=None,
            checkpoint_every=1,
            smaller_is_better=True,
            optimize_first=False,
            enable_amp=False,
            detect_anomalies=False
            ):
        """
        Starts the training run.

        Args:
            model:
            train_loader:
            val_loader:
            loggers: list of loggers that subscribe to various logging events
            max_epochs:
            use_gpu:
            gpu_id:
            resume_checkpoint: path to checkpoint to resume training from
            save_path: path to directory that blowtorch will save logs and checkpoints to
            run_name: name associated with this run, will be randomly created if None
            optimize_metric: train metric that will be used for optimization, will pick the first returned one if None
            checkpoint_metric: validation metric that will be used for checkpointing, will pick the first returned one if None
            checkpoint_every: every checkpoint_every epochs a checkpoint is saved, disregarding performance of the
            current model. This way it's always possible to resume the run from the latest (or near-latest) state
            smaller_is_better: ``True`` if we want to minimize, ``False`` if maximize
            optimize_first: whether optimization should occur during the first epoch
            enable_amp:
            detect_anomalies: enable autograd anomaly detection
        """
        self._model = model
        self._max_epochs = max_epochs
        self._use_gpu = use_gpu
        self._gpu_id = gpu_id
        self._resume_checkpoint = resume_checkpoint
        self._run_name = run_name
        self._save_path = save_path
        self._optimize_metric = optimize_metric
        self._checkpoint_metric = checkpoint_metric
        self._checkpoint_every = checkpoint_every
        self._smaller_is_better = smaller_is_better
        self._optimize_first = optimize_first
        self._enable_amp = enable_amp

        self._save_path = Path(self._save_path)
        self._save_path.mkdir(parents=True, exist_ok=True)

        # assign new random.Random() instance to coolname, such that slugs are different even though we have seeded
        replace_random(random.Random())

        if self._resume_checkpoint:
            if self._run_name is not None:
                raise ValueError('A run name cannot be specified when resuming from a previous run.')
            self._resume_checkpoint = Path(self._resume_checkpoint)
            if not self._resume_checkpoint.is_dir() or not (self._resume_checkpoint / 'checkpoints').exists():
                raise ValueError('Path to resume from should be the parent directory of the "checkpoints" folder.')
            self._run_name = self._resume_checkpoint.stem.split('_')[-1]
            self._save_path = self._resume_checkpoint
        else:
            if self._run_name is None:
                self._run_name = generate_slug(2)
                assert '_' not in self._run_name
            elif '_' in self._run_name:
                raise ValueError('Run name cannot contain "_".')
            # append consecutive number to run name
            self._run_name += f'-{get_highest_run(self._save_path) + 1}'
            self._save_path = self._save_path / (datetime.now().strftime("%y-%m-%d_%H-%M-%S") + '_' + self._run_name)
            self._save_path.mkdir(parents=True, exist_ok=False)

        checkpoints_path = self._save_path / 'checkpoints'
        checkpoints_path.mkdir(exist_ok=True)

        if loggers is None:
            loggers = []
        if not isinstance(loggers, (list, tuple)):
            loggers = [loggers]
        self._logger = LoggerSet([StandardLogger()] + loggers)
        self._logger.setup(self._save_path, self._run_name, self._resume_checkpoint is not None)

        if 'train_step' in self._bound_functions and 'val_step' in self._bound_functions:
            mode = 'step'
        elif 'train_epoch' in self._bound_functions and 'val_epoch' in self._bound_functions:
            mode = 'epoch'
        else:
            err_str = 'Need to register either both train_step/val_step or both train_epoch/val_epoch functions.'
            writer.error(err_str)
            raise ValueError(err_str)

        # Backend initialization
        try:
            if self._use_gpu:
                self._backend = GPUBackend(self._gpu_id, self._enable_amp)
            else:
                self._backend = CPUBackend()
        except Exception as e:
            writer.error(str(e))
            raise

        # load checkpoint if requested
        checkpoint = None
        start_epoch = 0
        if self._resume_checkpoint:
            writer.info(f'Resuming training from checkpoint {self._resume_checkpoint}')
            checkpoint = torch.load(checkpoints_path / 'latest', map_location='cpu')
            start_epoch = checkpoint['next_epoch']

        # backend takes care of initializing model, optimizers and schedulers
        self._backend.setup(self._model, self._bound_functions['configure_optimizers'], checkpoint)
        writer.info(f'Using {self._backend.get_name()}')

        if not self._optimize_first and start_epoch == 0:
            writer.info('Not optimizing during first epoch')

        self._logger.before_training_start(self._config.get_raw_config(), self._model, self._bound_functions)

        best_val = float('inf') if self._smaller_is_better else 0.

        for epoch in range(start_epoch, start_epoch + self._max_epochs):
            metrics = {}  # stores metrics of current epoch

            # train
            self._model.train()

            with writer.task(f'Training epoch {epoch}') as t:
                if mode == 'step':
                    step_metrics = []
                    for batch in t.tqdm(train_loader):
                        batch = self._backend.to_device(batch)

                        with torch.autograd.set_detect_anomaly(detect_anomalies):
                            # don't calculate grads if we're in epoch zero and not optimizing
                            torch.set_grad_enabled(self._optimize_first or epoch > 0)

                            train_metrics = self._backend.train_step(
                                self._bound_functions['train_step'],
                                batch=batch,
                                model=self._model,
                                is_validate=False,
                                device=self._backend.device,
                                epoch=epoch
                            )
                            assert isinstance(train_metrics, dict), '"train_step" should return a metrics dict.'

                            if self._optimize_metric is None:
                                metric = list(train_metrics.keys())[0]  # TODO possibility to state which one to optimize
                                writer.info(f'Selected metric "{metric}" for minimization')
                                self._optimize_metric = metric

                            if self._optimize_first or epoch > 0:
                                self._backend.optim_step(train_metrics[self._optimize_metric])

                        t.set_current_metrics({
                            self._optimize_metric: std_round(train_metrics[self._optimize_metric].item())})
                        step_metrics.append({k: float(v) for k, v in train_metrics.items()})

                        if 'after_train_step' in self._bound_functions:
                            self._bound_functions['after_train_step'](
                                model=self._model,
                                is_validate=False,
                                device=self._backend.device,
                                epoch=epoch
                            )

                    # calculate mean metrics
                    metrics['train'] = {
                        metric: np.array([dic[metric] for dic in step_metrics]).mean() for metric in step_metrics[0]
                    }

                elif mode == 'epoch':
                    # train_metrics = self._bound_functions['train_epoch'](data_loader=train_loader, model=self._model,
                    #                                                    is_validate=False, optimizers=self._optimizers)
                    # assert isinstance(train_metrics, dict), '"train_epoch" should return a metrics dict.'
                    # metrics['train'] = {k: float(v) for k, v in train_metrics.items()}
                    raise NotImplementedError()  # TODO wrap dataloader generator ?

                self._logger.after_pass(metrics['train'], epoch, is_validate=False)

                status_str = f'[Epoch {epoch} / Train] ' \
                             + ' '.join([f'{k}: {std_round(v)}' for k, v in metrics['train'].items()])
                t.success(status_str)

            # val
            self._model.eval()

            with writer.task(f'Validating epoch {epoch}') as t:
                if mode == 'step':
                    step_metrics = []
                    for batch in t.tqdm(val_loader):
                        batch = self._backend.to_device(batch)

                        val_metrics = self._backend.val_step(
                            self._bound_functions['val_step'],
                            batch=batch,
                            model=self._model,
                            is_validate=True,
                            device=self._backend.device,
                            epoch=epoch
                        )

                        assert isinstance(val_metrics, dict), '"val_step" should return a metrics dict.'
                        t.set_current_metrics({
                            self._optimize_metric: std_round(val_metrics[self._optimize_metric].item())})
                        step_metrics.append({k: float(v) for k, v in val_metrics.items()})

                    # calculate mean metrics
                    # todo agg function
                    metrics['val'] = {
                        metric: np.array([dic[metric] for dic in step_metrics]).mean() for metric in step_metrics[0]
                    }

                elif mode == 'epoch':
                    # val_metrics = self._bound_functions['val_epoch'](data_loader=val_loader, model=self._model,
                    #                                                  is_validate=True, optimizers=self._optimizers)
                    # assert isinstance(val_metrics, dict), '"val_epoch" should return a metrics dict.'
                    # metrics['val'] = {k: float(v) for k, v in val_metrics.items()}
                    raise NotImplementedError()

                # TODO specify metric to do scheduling on
                # if self._optimize_first is False, a warning will be raised by the schedulers which suggests that
                # optim.step() is called after scheduler.step(), which would normally result in the first epoch being
                # skipped from the learning rate scheduler. In our case, however, optim.step() was not called because
                # of self._optimize_first is False, and the epoch counter should indeed be increased.
                if epoch == 0 and not self._optimize_first:
                    warnings.simplefilter(action='ignore', category=UserWarning)
                    self._backend.scheduler_step(metrics['val'][self._optimize_metric])
                    warnings.filterwarnings('default')
                else:
                    self._backend.scheduler_step(metrics['val'][self._optimize_metric])

                self._logger.after_pass(metrics['val'], epoch, is_validate=True)

                status_str = f'[Epoch {epoch} / Val]   ' \
                             + ' '.join([f'{k}: {std_round(v)}' for k, v in metrics['val'].items()])
                t.success(status_str)

            if self._checkpoint_metric is None:
                metric = list(val_metrics.keys())[0]
                writer.info(f'Selected metric "{metric}" for checkpointing')
                self._checkpoint_metric = metric

            is_best = (self._smaller_is_better and metrics['val'][self._checkpoint_metric] < best_val) or \
                      (not self._smaller_is_better and metrics['val'][self._checkpoint_metric] > best_val)

            if is_best or epoch % self._checkpoint_every == 0:
                with writer.task(f'Saving checkpoint'):
                    checkpoint = {
                        'model': self._model.state_dict(),
                        'optimizers': {name: optim.state_dict() for name, optim in self._backend.optimizers.items()},
                        'schedulers': {name: sched.state_dict() for name, sched in self._backend.schedulers.items()},
                        'next_epoch': epoch + 1
                    }
                    path = checkpoints_path / f'epoch_{epoch}.pt'
                    torch.save(checkpoint, path)

                    latest_path = (checkpoints_path / 'latest')
                    best_path = (checkpoints_path / 'best')

                    if latest_path.is_symlink():
                        # delete previous latest checkpoint
                        checkpoint_file = latest_path.resolve()
                        latest_path.unlink()
                        if not (best_path.is_symlink() and best_path.resolve() == checkpoint_file):
                            # best_path symlink does not link to this checkpoint, so we can delete it
                            checkpoint_file.unlink()
                    # create new latest symlink
                    latest_path.symlink_to(path.name)

                    if is_best:
                        # delete old best checkpoint and symlink new one
                        if best_path.is_symlink():
                            checkpoint_file = best_path.resolve()
                            best_path.unlink()
                            checkpoint_file.unlink()
                        best_path.symlink_to(path.name)

                best_val = metrics['val'][self._checkpoint_metric]

        writer.success(f'Training finished')

    def add_config(self, path):
        self._config_files.append(path)

    def get_raw_config(self):
        return self._config.get_raw_config()

    def __getitem__(self, item):
        return self._config[item]

    @staticmethod
    def seed_all(seed):
        seed_all(seed)

    @staticmethod
    def set_deterministic(deterministic):
        set_deterministic(deterministic)

    @functools.wraps(run)
    def __call__(self, *args, **kwargs):
        self.run(*args, **kwargs)

    # DECORATORS #

    # TODO docstrings

    def init(self, f):
        self._bound_functions['init'] = f
        return make_wrapper(f)

    def train_step(self, f):
        self._bound_functions['train_step'] = f
        return make_wrapper(f)

    def after_train_step(self, f):
        self._bound_functions['after_train_step'] = f
        return make_wrapper(f)

    def validate_step(self, f):
        self._bound_functions['val_step'] = f
        return make_wrapper(f)

    def train_epoch(self, f):
        self._bound_functions['train_epoch'] = f
        return make_wrapper(f)

    def validate_epoch(self, f):
        self._bound_functions['val_epoch'] = f
        return make_wrapper(f)

    def configure_optimizers(self, f):
        self._bound_functions['configure_optimizers'] = f
        return make_wrapper(f)



import inspect
from abc import ABC, abstractmethod
from pathlib import Path
from typing import List

import yaml
from torch import nn

from .bound_functions import BoundFunctions
from . import _writer as writer


# TODO implement standardlogger and always use it


class BaseLogger(ABC):

    def __init__(self):
        pass

    @abstractmethod
    def setup(self, save_path: Path, run_name: str, resume: bool):
        """
        Used to setup the logger, given the run name
        """
        pass

    @abstractmethod
    def before_training_start(self, config: dict, model: nn.Module, bound_functions: BoundFunctions):
        pass

    @abstractmethod
    def after_pass(self, metrics: dict, epoch: int, is_validate: bool = False):
        pass

    @abstractmethod
    def after_epoch(self, epoch: int, model: nn.Module):
        pass


class StandardLogger(BaseLogger):

    def __init__(self):
        super().__init__()
        self._save_path = None
        self._log_file = None
        self._resume = None

    def setup(self, save_path: Path, run_name: str, resume: bool):
        self._save_path = save_path
        self._log_file = self._save_path / 'log.txt'
        self._resume = resume

    def before_training_start(self, config: dict, model: nn.Module, bound_functions: BoundFunctions):
        if self._resume:
            return

        # save config to file
        with (self._save_path / 'config.yaml').open('w') as fh:
            yaml.dump(config, fh)

        # save model summary and source to file
        with (self._save_path / 'model_summary.txt').open('w') as fh:
            fh.write(repr(model))

        with (self._save_path / 'source.txt').open('a') as fh:
            fh.write(inspect.getsource(type(model)))
            for function in bound_functions.values():
                fh.write('\n')
                fh.write(inspect.getsource(function))

    def after_pass(self, metrics: dict, epoch: int, is_validate: bool = False):
        status_str = f'[Epoch {epoch} / {"Val" if is_validate else "Train"}] ' \
                     + ' '.join([f'{k}: {v}' for k, v in metrics.items()]) + '\n'
        with self._log_file.open('a') as fh:
            fh.write(status_str)

    def after_epoch(self, epoch: int, model: nn.Module):
        pass


class WandbLogger(BaseLogger):

    def __init__(self, **kwargs):
        super().__init__()
        import wandb
        self._wandb = wandb
        self._wandb_args = kwargs
        self._resume = None

    def setup(self, save_path: Path, run_name: str, resume: bool):
        if 'name' in self._wandb_args:
            del self._wandb_args['name']
        self._resume = resume
        # always set resume=True, since every run is saved in its own directory anyway and wandb will just create a new
        # run if no previous run exists in the save_path directory. This way wandb automatically saves the run id in
        # the save_path directory and automatically resumes later, if desired.
        self._wandb.init(name=run_name, dir=save_path, resume=True, **self._wandb_args)

    def before_training_start(self, config: dict, model: nn.Module, bound_functions: BoundFunctions):
        if not self._resume:
            self._wandb.config.update(config)

    def after_pass(self, metrics: dict, epoch: int, is_validate: bool = False):
        prefix = 'val_' if is_validate else 'train_'
        self._wandb.log({(prefix + k): v for k, v in metrics.items()}, step=epoch)

    def after_epoch(self, epoch: int, model: nn.Module):
        pass


class TensorBoardLogger(BaseLogger):

    def __init__(self, log_dir=None, **kwargs):
        """
        Initializes a new TensorBoard logger.
        :param log_dir: create a subfolder for each run inside this directory. Defaults to run.save_path.
        :param **kwargs: arguments directly passed on to SummaryWriter.
        """
        super().__init__()
        # do this to fail early in case tensorboard is not installed
        from torch.utils.tensorboard import SummaryWriter
        self.summary_writer = None
        self.log_dir = log_dir
        self._kwargs = kwargs

    def setup(self, save_path: Path, run_name: str):
        if self.log_dir is None:
            # use blowtorch save_path
            log_dir = save_path
        else:
            log_dir = Path(self.log_dir) / save_path.name

        writer.info(f'Using {log_dir} as TensorBoard log_dir')
        from torch.utils.tensorboard import SummaryWriter
        self.summary_writer = SummaryWriter(str(log_dir), **self._kwargs)

    def before_training_start(self, config: dict, model: nn.Module, bound_functions: BoundFunctions):
        pass

    def after_pass(self, metrics: dict, epoch: int, is_validate: bool = False):
        postfix = '/val' if is_validate else '/train'
        for k, v in metrics.items():
            self.summary_writer.add_scalar(k + postfix, v, epoch)

    def after_epoch(self, epoch: int, model: nn.Module):
        pass


class LoggerSet(BaseLogger):

    def __init__(self, loggers: List[BaseLogger]):
        super().__init__()
        self._loggers = loggers

    def setup(self, save_path: Path, run_name: str, resume: bool):
        for logger in self._loggers:
            logger.setup(save_path, run_name, resume)

    def before_training_start(self, config: dict, model: nn.Module, bound_functions: BoundFunctions):
        for logger in self._loggers:
            logger.before_training_start(config, model, bound_functions)

    def after_pass(self, metrics: dict, epoch: int, is_validate: bool = False):
        for logger in self._loggers:
            logger.after_pass(metrics, epoch, is_validate)

    def after_epoch(self, epoch: int, model: nn.Module):
        for logger in self._loggers:
            logger.after_epoch(epoch, model)
